# Spec version 0.3.1
# This file is auto-generated by setup.py

from .thing_types import ThingTypesBase


class AnnotationTypesBase(ThingTypesBase):
    """
    Inherit from this class to build your own custom annotation
    vocabularies. Each member's name should be the name of the
    type, and each member's value should be the URI of that type.
    """
    ...


class AnnotationTypes(AnnotationTypesBase):
    """
    This enum contains the URIs for the MMIF annotation types defined in the spec version 0.3.1.
    Use this to quickly get the correct URIs for those types, or use the objects themselves in your
    code and they will serialize to the strings.
    """
    Annotation = 'http://mmif.clams.ai/0.3.1/vocabulary/Annotation'
    Region = 'http://mmif.clams.ai/0.3.1/vocabulary/Region'
    TimePoint = 'http://mmif.clams.ai/0.3.1/vocabulary/TimePoint'
    Interval = 'http://mmif.clams.ai/0.3.1/vocabulary/Interval'
    Span = 'http://mmif.clams.ai/0.3.1/vocabulary/Span'
    TimeFrame = 'http://mmif.clams.ai/0.3.1/vocabulary/TimeFrame'
    Chapter = 'http://mmif.clams.ai/0.3.1/vocabulary/Chapter'
    Polygon = 'http://mmif.clams.ai/0.3.1/vocabulary/Polygon'
    BoundingBox = 'http://mmif.clams.ai/0.3.1/vocabulary/BoundingBox'
    VideoObject = 'http://mmif.clams.ai/0.3.1/vocabulary/VideoObject'
    Relation = 'http://mmif.clams.ai/0.3.1/vocabulary/Relation'
    Alignment = 'http://mmif.clams.ai/0.3.1/vocabulary/Alignment'



from .annotation import __all__ as anno_all
from .annotation import *
from .mmif import __all__ as mmif_all
from .mmif import *
from .view import __all__ as view_all
from .view import *

__all__ = anno_all + mmif_all + view_all



from .thing_types import ThingTypesBase
from .thing_types import ThingType
from .annotation_types import AnnotationTypesBase
from .annotation_types import AnnotationTypes
from .document_types import DocumentTypesBase
from .document_types import DocumentTypes



from mmif.ver import __version__
from mmif.ver import __specver__
from mmif.vocabulary import *
from mmif.serialize import *

_res_pkg = 'res'
_ver_pkg = 'ver'
_vocabulary_pkg = 'vocabulary'
_schema_res_name = 'mmif.json'

# This package file used to have several importlib tricks
# in an attempt to enable re-use of some variable (mostly subpackage names)
# in `setup.py` by importing this.
# But I dropped that idea as it introduced hard-to-solve circular dependency problems.


"""
The :mod:`view` module contains the classes used to represent a MMIF view
as a live Python object.

In MMIF, views are created by apps in a pipeline that are annotating
data that was previously present in the MMIF file.
"""
from datetime import datetime
from typing import Dict, Union, Optional, Generator, List, cast
import dateutil.parser
from pyrsistent import pmap, pvector

from .annotation import Annotation, Document
from .model import FreezableMmifObject, FreezableDataList, FreezableDataDict
from mmif.vocabulary import ThingTypesBase


__all__ = ['View', 'ViewMetadata', 'Contain']


class View(FreezableMmifObject):
    """
    View object that represents a single view in a MMIF file.

    A view is identified by an ID, and contains certain metadata,
    a list of annotations, and potentially a JSON-LD ``@context``
    IRI.

    If ``view_obj`` is not provided, an empty View will be generated.

    :param view_obj: the JSON data that defines the view
    """

    def __init__(self, view_obj: Union[bytes, str, dict] = None) -> None:
        self.id: str = ''
        self.metadata: ViewMetadata = ViewMetadata()
        self.annotations: AnnotationsList = AnnotationsList()
        self.disallow_additional_properties()
        self._attribute_classes = pmap({
            'metadata': ViewMetadata,
            'annotations': AnnotationsList
        })
        self._required_attributes = pvector(["id", "metadata", "annotations"])
        super().__init__(view_obj)
        for item in self.annotations:
            if isinstance(item, Document):
                item.parent = self.id

    def new_contain(self, at_type: Union[str, ThingTypesBase], contain_dict: dict = None) -> Optional['Contain']:
        """
        Adds a new element to the ``contains`` metadata.

        :param at_type: the ``@type`` of the annotation type being added
        :param contain_dict: any metadata associated with the annotation type
        :return: the generated :class:`Contain` object
        """
        return self.metadata.new_contain(at_type, contain_dict)

    def new_annotation(self, aid: str, at_type: Union[str, ThingTypesBase], overwrite=False) -> 'Annotation':
        """
        Generates a new :class:`mmif.serialize.annotation.Annotation`
        object and adds it to the current view.

        Fails if there is already an annotation with the same ID
        in the view, unless ``overwrite`` is set to True.

        :param aid: the desired ID of the annotation
        :param at_type: the desired ``@type`` of the annotation
        :param overwrite: if set to True, will overwrite an
                          existing annotation with the same ID
        :raises KeyError: if ``overwrite`` is set to False and
                          an annotation with the same ID exists
                          in the view
        :return: the generated :class:`mmif.serialize.annotation.Annotation`
        """
        new_annotation = Annotation()
        new_annotation.at_type = at_type
        new_annotation.id = aid
        return self.add_annotation(new_annotation, overwrite)

    def add_annotation(self, annotation: 'Annotation', overwrite=False) -> 'Annotation':
        """
        Adds an annotation to the current view.

        Fails if there is already an annotation with the same ID
        in the view, unless ``overwrite`` is set to True.

        :param annotation: the :class:`mmif.serialize.annotation.Annotation`
                           object to add
        :param overwrite: if set to True, will overwrite an
                          existing annotation with the same ID
        :raises KeyError: if ``overwrite`` is set to False and
                          an annotation with the same ID exists
                          in the view
        :return: the same Annotation object passed in as ``annotation``
        """
        self.annotations.append(annotation, overwrite)
        self.new_contain(annotation.at_type)
        return annotation

    def add_document(self, document: Document, overwrite=False) -> Annotation:
        """
        Appends a Document object to the annotations list.

        Fails if there is already a document with the same ID in the annotations list.

        :param document: the Document object to add
        :param overwrite: if set to True, will overwrite
                          an existing view with the same ID
        :return: None
        """
        document.parent = self.id
        return self.add_annotation(document, overwrite)

    def get_annotations(self, at_type: Union[str, ThingTypesBase] = None, **properties) -> Generator[Annotation, None, None]:
        """
        Look for certain annotations in this view, specified by parameters

        :param at_type: @type of the annotations to look for. When this is None, any @type will match.
        :param properties: properties of the annotations to look for. When given more than one property, all properties must match. Note that some annotation properties (annotation type medata) are specified in the `contains` view metadata
        """
        def prop_check(k, v, *props):
            return any(k in prop and prop[k] == v for prop in props)

        if at_type:
            at_type_metadata = self.metadata.contains[str(at_type)]
        for annotation in self.annotations:
            if not at_type:
                if all(map(lambda kv: prop_check(kv[0], kv[1], annotation.properties, self.metadata.contains[str(annotation.at_type)]), properties.items())):
                    yield annotation
            if at_type and str(annotation.at_type) == str(at_type):
                if all(map(lambda kv: prop_check(kv[0], kv[1], annotation.properties, at_type_metadata), properties.items())):
                    yield annotation

    def get_documents(self) -> List[Document]:
        return [cast(Document, annotation) for annotation in self.annotations if annotation.is_document()]

    def get_document_by_id(self, doc_id) -> Document:
        doc_found = self.annotations.get(doc_id)
        if doc_found is None or not isinstance(doc_found, Document):
            raise KeyError(f"{doc_id} not found in view {self.id}.")
        else:
            return doc_found

    def __getitem__(self, key: str) -> 'Annotation':
        """
        getitem implementation for View.

        >>> obj = View('''{"id": "v1","metadata": {"contains": {"BoundingBox": {"unit": "pixels"}},"document": "m1","tool": "http://tools.clams.io/east/1.0.4"},"annotations": [{"@type": "BoundingBox","properties": {"id": "bb1","coordinates": [[90,40], [110,40], [90,50], [110,50]] }}]}''')
        >>> type(obj['bb1'])
        <class 'mmif.serialize.annotation.Annotation'>
        >>> obj['asdf']
        Traceback (most recent call last):
            ...
        KeyError: 'Annotation ID not found: asdf'

        :raises KeyError: if the key is not found
        :param key: the search string.
        :return: the :class:`mmif.serialize.annotation.Annotation` object searched for
        """
        if key in self._named_attributes():
            return self.__dict__[key]
        anno_result = self.annotations.get(key)
        if not anno_result:
            raise KeyError("Annotation ID not found: %s" % key)
        return anno_result
    
    def set_error(self, err_message: str, err_trace: str) -> None:
        self.metadata.set_error(err_message, err_trace)
        self.annotations.empty()


class ViewMetadata(FreezableMmifObject):
    """
    ViewMetadata object that represents the ``metadata`` object within a MMIF view.

    :param viewmetadata_obj: the JSON data that defines the metadata
    """

    def __init__(self, viewmetadata_obj: Union[bytes, str, dict] = None) -> None:
        self.document: str = ''
        self.timestamp: Optional[datetime] = None
        self.app: str = ''
        self.contains: ContainsDict = ContainsDict()
        self.parameters: dict = {}
        self.error: dict = {}
        self._required_attributes = pvector(["app"])
        # in theory, either `contains` or `error` should appear in a `view`
        # but with current implementation, there's no easy way to set a condition 
        # for `oneOf` requirement 
        # see MmifObject::_required_attributes in model.py 
        super().__init__(viewmetadata_obj)

    def _deserialize(self, input_dict: dict) -> None:
        """
        Extends base ``_deserialize`` method to initialize
        ``contains`` as a dict of Contain objects.

        :param input_dict: the JSON data that defines the metadata
        :return: None
        """
        try:
            self.contains = ContainsDict(input_dict.pop('contains'))
        except KeyError:
            # means input_dict don't have `contains`, so we'll leave it empty
            pass
        super()._deserialize(input_dict)

    def _find_match_hotfix(self, key: str) -> bool:
        """
        Checks the existing types in the contains dict to see if
        the type passed in as ``key`` has the same shortname.

        FIXME: this will produce undesired results if there is a
         shortname conflict in the view.

        :param key: the type (shortname or IRI) to check
        :return: whether ``key`` already has a match in the ``contains`` dict
        """
        exists = False
        for existing_type in self.contains.keys():
            if key.split('/')[-1] == existing_type.split('/')[-1]:
                exists = True
                break
        return exists

    def new_contain(self, at_type: Union[str, ThingTypesBase], contain_dict: dict = None) -> Optional['Contain']:
        """
        Adds a new element to the ``contains`` dictionary.

        :param at_type: the ``@type`` of the annotation type being added
        :param contain_dict: any metadata associated with the annotation type
        :return: the generated :class:`Contain` object
        """
        if isinstance(at_type, ThingTypesBase):
            exists = self._find_match_hotfix(at_type.name) or self._find_match_hotfix(at_type.value)
            final_key = at_type.value
        else:
            exists = self._find_match_hotfix(at_type)
            final_key = at_type

        if not exists:
            new_contain = Contain(contain_dict)
            self.contains[final_key] = new_contain
            return new_contain

    def add_parameters(self, param_dict: dict = None, **param_kwargs):
        if param_dict is None:
            self.parameters = {}
        else:
            self.parameters = param_dict
        self.parameters.update(dict(param_kwargs))

    def add_parameter(self, param_key, param_value):
        self.parameters[param_key] = param_value

    def get_parameter(self, param_key):
        try:
            return self.parameters[param_key]
        except KeyError:
            raise KeyError(f"parameter \"{param_key}\" is not set in the view: {self.serialize()}")
    
    def set_error(self, message: str, stack_trace: str):
        self.error = ErrorDict({"message": message, "stackTrace": stack_trace})
        self.contains = {}


class ErrorDict(FreezableMmifObject):
    """
    Error object that stores information about error occurred during processing. 
    """
    def __init__(self, error_obj: Union[bytes, str, dict] = None) -> None:
        self.message: str = ''
        self.stackTrace: str = ''
        super().__init__(error_obj)
        

class Contain(FreezableMmifObject):
    """
    Contain object that represents the metadata of a single
    annotation type in the ``contains`` metadata of a MMIF view.

    :param contain_obj: the metadata that defines this object
    """

    def __init__(self, contain_obj: Union[bytes, str, dict] = None) -> None:
        # TODO (krim @ 8/19/20): rename `producer` to `app` maybe?
        self.producer: str = ''
        self.gen_time: Optional[datetime] = None
        super().__init__(contain_obj)

    def _deserialize(self, input_dict: dict) -> None:
        """
        Extends base ``_deserialize`` method to initialize the
        ``gen_time`` metadata as a :class:`datetime.datetime` object.

        :param input_dict: the metadata that defines this object
        :return: None
        """
        super()._deserialize(input_dict)
        if 'gen_time' in self.__dict__ and isinstance(self.gen_time, str):
            self.gen_time = dateutil.parser.isoparse(self.gen_time)


class AnnotationsList(FreezableDataList[Union[Annotation, Document]]):
    """
    AnnotationsList object that implements :class:`mmif.serialize.model.DataList`
    for :class:`mmif.serialize.annotation.Annotation`.
    """
    _items: Dict[str, Union[Annotation, Document]]

    def _deserialize(self, input_list: list) -> None:
        """
        Extends base ``_deserialize`` method to initialize ``items`` as a dict from
        annotation IDs to :class:`mmif.serialize.annotation.Annotation` objects.

        :param input_list: the JSON data that defines the list of annotations
        :return: None
        """
        self._items = {item['properties']['id']: Document(item)
                       if item['_type'].endswith("Document") else Annotation(item)
                       for item in input_list}

    def append(self, value: Union[Annotation, Document], overwrite=False) -> None:
        """
        Appends an annotation to the list.

        Fails if there is already an annotation with the same ID
        in the list, unless ``overwrite`` is set to True.

        :param value: the :class:`mmif.serialize.annotation.Annotation`
                      object to add
        :param overwrite: if set to True, will overwrite an
                          existing annotation with the same ID
        :raises KeyError: if ``overwrite`` is set to False and
                          an annotation with the same ID exists
                          in the list
        :return: None
        """
        super()._append_with_key(value.id, value, overwrite)
    
    def empty(self):
        self._items = {}


class ContainsDict(FreezableDataDict[Contain]):
    _items: Dict[str, Contain]

    def _deserialize(self, input_dict: dict) -> None:
        self._items = {key: Contain(value) for key, value in input_dict.items()}

    def update(self, other: Union[dict, 'ContainsDict'], overwrite=False):
        for k, v in other.items():
            self._append_with_key(k, v, overwrite=overwrite)



# This file is auto-generated by setup.py

from enum import Enum


class ThingTypesBase(Enum):
    def _serialize(self):
        return str(self)

    def __str__(self):
        return self.value


class ThingType(ThingTypesBase):
    """
    This enum contains the URI for the MMIF Thing types defined in the spec version 0.3.1.
    Use this to quickly get the correct URI for this type, or use the object itself in your
    code and it will serialize to the string.
    """
    Thing = 'http://mmif.clams.ai/0.3.1/vocabulary/Thing'



"""
The :mod:`model` module contains the classes used to represent an
abstract MMIF object as a live Python object.

The :class:`MmifObject` class or one of its derivatives is subclassed by
all other classes defined in this SDK, except for :class:`MmifObjectEncoder`.

These objects are generally instantiated from JSON, either as a string
or as an already-loaded Python dictionary. This base class provides the
core functionality for deserializing MMIF JSON data into live objects
and serializing live objects into MMIF JSON data. Specialized behavior
for the different components of MMIF is added in the subclasses.
"""

import logging
import json
from pyrsistent import pvector, m, pmap, s, PVector, PMap, PSet, thaw
from datetime import datetime

from deepdiff import DeepDiff
from typing import Union, Any, Dict, Optional, TypeVar, Generic, Type, Generator, Iterator, List

T = TypeVar('T')

__all__ = [
    'MmifObject',
    'FreezableMmifObject',
    'MmifObjectEncoder',
    'DataList',
    'DataDict',
    'FreezableDataList',
    'FreezableDataDict'
]


class MmifObject(object):
    """
    Abstract superclass for MMIF related key-value pair objects.

    Any MMIF object can be initialized as an empty placeholder or
    an actual representation with a JSON formatted string or equivalent
    `dict` object argument.

    This superclass has three specially designed instance variables, and these
    variable names cannot be used as attribute names for MMIF objects.

    1. _unnamed_attributes:
       Only can be either None or an empty dictionary. If it's set to None,
       it means the class won't take any ``Additional Attributes`` in the JSON
       schema sense. If it's a dict, users can throw any k-v pairs to the
       class, EXCEPT for the reserved two key names.
    2. _attribute_classes:
       This is a dict from a key name to a specific python class to use for
       deserialize the value. Note that a key name in this dict does NOT
       have to be a *named* attribute, but is recommended to be one.
    3. _required_attributes:
       This is a simple list of names of attributes that are required in the object.
       When serialize, an object will skip its *empty* (e.g. zero-length, or None)
       attributes unless they are in this list. Otherwise, the serialized JSON
       string would have empty representations (e.g. ``""``, ``[]``).

    # TODO (krim @ 8/17/20): this dict is however, a duplicate with the type hints in the class definition.
    Maybe there is a better way to utilize type hints (e.g. getting them as a programmatically), but for now
    developers should be careful to add types to hints as well as to this dict.

    Also note that those two special attributes MUST be set in the __init__()
    before calling super method, otherwise deserialization will not work.

    And also, a subclass that has one or more *named* attributes, it must
    set those attributes in the __init__() before calling super method. When
    serializing a MmifObject, all *empty* attributes will be ignored, so for
    optional named attributes, you must leave the values empty (len == 0), but
    NOT None. Any None-valued named attributes will cause issues with current
    implementation.

    :param mmif_obj: JSON string or `dict` to initialize an object.
     If not given, an empty object will be initialized, sometimes with
     an ID value automatically generated, based on its parent object.
    """
    
    reserved_names: PSet = s('reserved_names',
                             '_unnamed_attributes',
                             '_attribute_classes',
                             '_required_attributes',
                             # used in freezable subclasses
                             '_frozen',
                             # used in Document class to store parent view id
                             '_parent_view_id')
    _unnamed_attributes: Optional[dict]
    _attribute_classes: PMap = m()  # Mapping: str -> Type
    _required_attributes: PVector

    def __init__(self, mmif_obj: Union[bytes, str, dict] = None) -> None:
        if isinstance(mmif_obj, bytes):
            mmif_obj = mmif_obj.decode('utf8')
        if not hasattr(self, '_required_attributes'):
            self._required_attributes = pvector()
        if not hasattr(self, '_unnamed_attributes'):
            self._unnamed_attributes = {}
        if mmif_obj is not None:
            self.deserialize(mmif_obj)

    def disallow_additional_properties(self) -> None:
        """
        Call this method in :func:`__init__` to prevent the insertion
        of unnamed attributes after initialization.
        """
        self._unnamed_attributes = None

    def set_additional_property(self, key: str, value: Any) -> None:
        """
        Method to set values in _unnamed_attributes.

        :param key: the attribute name
        :param value: the desired value
        :return: None
        :raise: AttributeError if additional properties are disallowed by :func:`disallow_additional_properties`
        """
        if self._unnamed_attributes is None:
            raise AttributeError(f"Additional properties are disallowed by {self.__class__}")
        self._unnamed_attributes[key] = value # pytype: disable=unsupported-operands

    def _named_attributes(self) -> Generator[str, None, None]:
        """
        Returns a generator of the names of all of this object's named attributes.

        :return: generator of names of all named attributes
        """
        return (n for n in self.__dict__.keys() if n not in self.reserved_names)

    def serialize(self, pretty: bool = False) -> str:
        """
        Generates JSON representation of an object.

        :param pretty: If True, returns string representation with indentation.
        :return: JSON string of the object.
        """
        return json.dumps(self._serialize(), indent=2 if pretty else None, cls=MmifObjectEncoder)

    def _serialize(self, alt_container: Dict = None) -> dict:
        """
        Maps a MMIF object to a plain python dict object,
        rewriting internal keys that start with '_' to
        start with '@' per the JSON-LD schema.

        If a subclass needs special treatment during the mapping, it needs to
        override this method.

        :return: the prepared dictionary
        """
        container = alt_container if alt_container is not None else self._unnamed_attributes
        serializing_obj = {}
        try:
            for k, v in container.items():   # pytype: disable=attribute-error
                if v is None:
                    continue
                if isinstance(v, (PSet, PVector, PMap)):
                    v = thaw(v)
                if k.startswith('_'):   # _ as a placeholder ``@`` in json-ld
                    k = f'@{k[1:]}'
                serializing_obj[k] = v
        except AttributeError as e:
            # means _unnamed_attributes is None, so nothing unnamed would be serialized
            pass
        for k, v in self.__dict__.items():
            if k in self.reserved_names:
                continue
            if k not in self._required_attributes and self.is_empty(v):
                continue
            if isinstance(v, (PSet, PVector, PMap)):
                v = thaw(v)
            if k.startswith('_'):       # _ as a placeholder ``@`` in json-ld
                k = f'@{k[1:]}'
            serializing_obj[k] = v
        return serializing_obj

    @staticmethod
    def is_empty(obj) -> bool:
        """
        return True if the obj is None or "emtpy". The emptiness first defined as
        having zero length. But for objects that lack __len__ method, we need
        additional check.
        """
        if obj is None:
            return True
        if hasattr(obj, '__len__') and len(obj) == 0:
            return True
        return False

    @staticmethod
    def _load_json(json_obj: Union[dict, str]) -> dict:
        """
        Maps JSON-format MMIF strings and dicts into Python dicts
        with identifier-compliant keys. To do this, it replaces "@"
        signs in JSON-LD field names with "_" to be python-compliant.

        >>> "_type" in MmifObject._load_json('{ "@type": "some_type", "@value": "some_value"}').keys()
        True
        >>> "_value" in MmifObject._load_json('{ "@type": "some_type", "@value": "some_value"}').keys()
        True

        :param json_str: the JSON data to load and process
        :return: the mapped data as a dict
        """
        def from_atsign(d: Dict[str, Any]) -> dict:
            for k in list(d.keys()):
                if k.startswith('@'):
                    d[f'_{k[1:]}'] = d.pop(k)
            return d

        def deep_from_atsign(d: dict) -> dict:
            new_d = d.copy()
            from_atsign(new_d)
            for key, value in new_d.items():
                if type(value) is dict:
                    new_d[key] = deep_from_atsign(value)
            return new_d

        if type(json_obj) is dict:
            return deep_from_atsign(json_obj)
        elif type(json_obj) is str:
            return json.loads(json_obj, object_hook=from_atsign)
        else:
            raise TypeError(f"tried to load MMIF JSON in a format other than str or dict: {type(json_obj)}")

    def deserialize(self, mmif_json: Union[str, dict]) -> None:
        """
        Takes a JSON-formatted string or a simple `dict` that's json-loaded from
        such a string as an input and populates object's fields with the values
        specified in the input.

        :param mmif_json: JSON-formatted string or dict from such a string
         that represents a MMIF object
        """
        mmif_json = self._load_json(mmif_json)
        self._deserialize(mmif_json)

    def _deserialize(self, input_dict: dict) -> None:
        """
        Maps a plain python dict object to a MMIF object.
        If a subclass needs a special treatment during the mapping, it needs to
        override this method.

        This default method won't work for generic types (e.g. List[X], Dict[X, Y]).
        For now, lists are abstracted as DataList and dicts are abstracted as XXXMedata classes.
        However, if an attribute uses a generic type (e.g. view_metadata.contains: Dict[str, Contain])
        that class should override _deserialize of its own.

        :param input_dict: the prepared JSON data that defines the object
        """
        for k, v in input_dict.items():
            if self._attribute_classes and k in self._attribute_classes:
                self[k] = self._attribute_classes[k](v)
            else:
                self[k] = v

    def __str__(self) -> str:
        return self.serialize(False)

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self)) and \
               len(DeepDiff(self, other, ignore_order=True, report_repetition=True, exclude_types=[datetime])) == 0

    def __len__(self) -> int:
        return sum([not self.is_empty(self[named]) for named in self._named_attributes()]) \
               + (len(self._unnamed_attributes) if self._unnamed_attributes else 0)

    def __setitem__(self, key, value) -> None:
        if key in self.reserved_names:
            raise KeyError("can't set item on a reserved name")
        if key in self._named_attributes():
            if self._attribute_classes and key in self._attribute_classes \
                    and not isinstance(value, (self._attribute_classes[key])):
                self.__dict__[key] = self._attribute_classes[key](value)
            else:
                self.__dict__[key] = value
        else:
            if self._attribute_classes and key in self._attribute_classes \
                    and not isinstance(value, (self._attribute_classes[key])):
                self.set_additional_property(key, self._attribute_classes[key](value))
            else:
                self.set_additional_property(key, value)

    def __contains__(self, key: str) -> bool:
        try:
            self.__getitem__(key)
            return True
        except (TypeError, AttributeError, KeyError):
            return False

    def __getitem__(self, key) -> Union['MmifObject', str, datetime]:
        if key in self._named_attributes():
            return self.__dict__[key]
        if self._unnamed_attributes is None:
            raise AttributeError(f"Additional properties are disallowed by {self.__class__}")
        return self._unnamed_attributes[key]


class FreezableMmifObject(MmifObject):

    def __init__(self, *args, **kwargs) -> None:
        self._frozen = False
        super().__init__(*args, **kwargs)

    def is_frozen(self):
        return self._frozen

    def freeze(self) -> None:
        """
        Shallowly freezes this FreezableMmifObject, preventing attribute assignments with `=`.
        Makes no promises about the mutability of state within the object, only the references
        to that state.

        :return: asdf
        """
        self._frozen = True

    def deep_freeze(self, *additional_containers: str) -> bool:
        """
        Deeply freezes this FreezableMmifObject, calling deep_freeze on all FreezableMmifObjects
        contained as attributes or members of iterable attributes.

        Note: in general, this makes no promises about the mutability of non-FreezableMmifObject
        state within the object. However, if all attributes and members of iterable attributes
        are either Freezable or hashable, this method will return True. Note that whether an object
        is hashable is not a contract of immutability but merely a suggestion, as anyone can
        implement __hash__.

        :param additional_containers: any names of attributes in the object that should have
                                      their contents frozen but not themselves. This is only
                                      used for FreezableDataList and FreezableDataDict classes
                                      to freeze their contents.
        :return: True if all state is either Freezable or Hashable
        """
        fully_frozen = True

        def _pyrsist(element):
            nonlocal fully_frozen

            if isinstance(element, (list, PVector)):
                return pvector(_pyrsist(item) for item in element)
            elif isinstance(element, (dict, PMap)):
                return pmap({key: _pyrsist(value) for key, value in element.items()})
            elif isinstance(element, FreezableMmifObject):
                fully_frozen &= element.deep_freeze()
                return element
            elif element is not None and (not hasattr(element, '__hash__')
                                          or element.__class__.__hash__ in {object.__hash__, None}):
                # element is most likely mutable and not freezable
                fully_frozen = False
                return element
            else:
                # element is most likely immutable
                return element

        # freeze unnamed attributes if there are any
        if hasattr(self, '_unnamed_attributes') and self._unnamed_attributes is not None:
            self._unnamed_attributes = _pyrsist(self._unnamed_attributes)

        # freeze named attributes
        for name in self._named_attributes():
            self.__setattr__(name, _pyrsist(self.__getattribute__(name)))

        # freeze additional containers passed in (currently only used for DataLists and DataDicts
        # to freeze contents of _items without destroying insertion order by converting to a PMap)
        for name in additional_containers:
            container = self.__getattribute__(name)
            if isinstance(container, dict):
                iter_pairs = container.items()
            elif isinstance(container, list):
                iter_pairs = enumerate(container)
            else:
                raise ValueError("additional_containers should only be of types dict or list")
            for key, value in iter_pairs:
                container[key] = _pyrsist(value)

        self.freeze()

        return fully_frozen

    def __setattr__(self, name, value) -> None:
        """
        Overrides object.__setattr__(self, name, value) to prevent
        attribute assignment if the object has been frozen.

        :param name: the attribute name
        :param value: the desired value
        """
        if '_frozen' not in self.__dict__ or not self._frozen:
            object.__setattr__(self, name, value)
        else:
            raise TypeError("frozen FreezableMmifObject should be immutable")

    def __setitem__(self, key, value):
        """
        Overrides the __setitem__ method of  to
        prevent item assignment if the object has been frozen.

        :param key: t
        :param value:
        :return:
        """
        if '_frozen' not in self.__dict__ or not self._frozen:
            setitem = super().__setitem__
            if logging.getLogger().level == logging.DEBUG:
                logging.debug(setitem)
            setitem(key, value)
        else:
            raise TypeError("frozen FreezableMmifObject should be immutable")


class MmifObjectEncoder(json.JSONEncoder):
    """
    Encoder class to define behaviors of de-/serialization
    """

    def default(self, obj: 'MmifObject'):
        """
        Overrides default encoding behavior to prioritize :func:`MmifObject.serialize()`.
        """
        if hasattr(obj, '_serialize'):
            return obj._serialize()
        elif hasattr(obj, 'isoformat'):         # for datetime objects
            return obj.isoformat()
        elif hasattr(obj, '__str__'):
            return str(obj)
        else:
            return json.JSONEncoder.default(self, obj)


class DataList(MmifObject, Generic[T]):
    """
    The DataList class is an abstraction that represents the
    various lists found in a MMIF file, such as documents, subdocuments,
    views, and annotations.

    :param Union[str, list] mmif_obj: the data that the list contains
    """
    def __init__(self, mmif_obj: Union[bytes, str, list] = None):
        self.reserved_names = self.reserved_names.add('_items')
        self._items: Dict[str, T] = dict()
        self.disallow_additional_properties()
        if mmif_obj is None:
            mmif_obj = []
        super().__init__(mmif_obj)

    def _serialize(self, *args, **kwargs) -> list:
        """
        Internal serialization method. Returns a list.

        :return: list of the values of the internal dictionary.
        """
        return list(super()._serialize(self._items).values())

    def deserialize(self, mmif_json: Union[str, list]) -> None:
        """
        Passes the input data into the internal deserializer.
        """
        super().deserialize(mmif_json)  # pytype: disable=unsupported-operands

    @staticmethod
    def _load_json(json_list: Union[list, str]) -> list:
        if type(json_list) is str:
            json_list = json.loads(json_list)
        return [MmifObject._load_json(obj) for obj in json_list]
    
    def _deserialize(self, input_list: dict) -> None:
        raise NotImplementedError()

    def get(self, key: str) -> Optional[T]:
        """
        Standard dictionary-style get() method, albeit with no ``default``
        parameter. Relies on the implementation of __getitem__.

        Will return ``None`` if the key is not found.

        :param key: the key to search for
        :return: the value matching that key
        """
        try:
            return self[key]
        except KeyError:
            return None

    def _append_with_key(self, key: str, value: T, overwrite=False) -> None:
        """
        Internal method for appending a key-value pair. Subclasses should
        implement an append() method that extracts a key from the list data
        or generates a key programmatically (such as an index), depending
        on the data type.

        :param key: the desired key to append
        :param value: the value associated with the key
        :param overwrite: if set to True, will overwrite an existing K-V pair
         if the key already exists. Otherwise, raises a KeyError.
        :raise KeyError: if ``overwrite`` is False and the ``key`` is already
         present in the DataList.
        :return: None
        """
        if not overwrite and key in self._items:
            raise KeyError(f"Key {key} already exists")
        else:
            self[key] = value

    def append(self, value, overwrite):
        raise NotImplementedError()

    def __getitem__(self, key: str) -> T:
        if key not in self.reserved_names:
            return self._items.__getitem__(key)
        else:
            raise KeyError("Don't use __getitem__ to access a reserved name")

    def __setitem__(self, key: str, value: T):
        if key not in self.reserved_names:
            self._items.__setitem__(key, value)
        else:
            super().__setitem__(key, value)

    def __iter__(self) -> Iterator[T]:
        return self._items.values().__iter__()

    def __len__(self) -> int:
        return self._items.__len__()

    def __reversed__(self) -> Iterator[T]:
        return reversed(list(self._items.values()))

    def __contains__(self, item) -> bool:
        return item in self._items


class FreezableDataList(FreezableMmifObject, DataList[T]):
    def _deserialize(self, input_dict: dict) -> None:
        raise NotImplementedError()

    def deep_freeze(self, *args, **kwargs) -> bool:
        return super().deep_freeze('_items')


class DataDict(MmifObject, Generic[T]):
    def __init__(self, mmif_obj: Union[bytes, str, dict] = None):
        self.reserved_names = self.reserved_names.add('_items')
        self._items: Dict[str, T] = dict()
        self.disallow_additional_properties()
        if mmif_obj is None:
            mmif_obj = {}
        super().__init__(mmif_obj)

    def _serialize(self, *args, **kwargs) -> dict:
        return super()._serialize(self._items)

    def _deserialize(self, input_dict: dict) -> None:
        raise NotImplementedError()

    def get(self, key: str) -> Optional[T]:
        return self._items.get(key)

    def _append_with_key(self, key: str, value: T, overwrite=False) -> None:
        if not overwrite and key in self._items:
            raise KeyError(f"Key {key} already exists")
        else:
            self[key] = value

    def update(self, other, overwrite):
        raise NotImplementedError()

    def items(self):
        return self._items.items()

    def keys(self):
        return self._items.keys()

    def values(self):
        return self._items.values()

    def __getitem__(self, key: str) -> T:
        if key not in self.reserved_names:
            return self._items.__getitem__(key)
        else:
            raise KeyError("Don't use __getitem__ to access a reserved name")

    def __setitem__(self, key: str, value: T):
        if key not in self.reserved_names:
            self._items.__setitem__(key, value)
        else:
            super().__setitem__(key, value)

    def __iter__(self):
        return self._items.__iter__()

    def __len__(self):
        return self._items.__len__()

    def __contains__(self, item):
        return item in self._items


class FreezableDataDict(FreezableMmifObject, DataDict[T]):
    def _deserialize(self, input_dict: dict) -> None:
        raise NotImplementedError()

    def deep_freeze(self, *args, **kwargs) -> bool:
        return super().deep_freeze('_items')
